@@ -4,6 +4,7 @@
 import re
 import unittest
 import jsbeautifier
+import six
 
 class TestJSBeautifier(unittest.TestCase):
     def test_unescape(self):
@@ -38,6 +39,10 @@ class TestJSBeautifier(unittest.TestCase):
         test_fragment = self.decodesto
         bt = self.bt
 
+        # unicode support
+        bt('var ' + six.unichr(3232) + '_' + six.unichr(3232) + ' = "hi";');
+        bt('var ' + six.unichr(228) + 'x = {\n    ' + six.unichr(228) + 'rgerlich: true\n};');
+
         bt('');
         bt('return .5');
         test_fragment('    return .5');
@@ -46,6 +51,10 @@ class TestJSBeautifier(unittest.TestCase):
         bt("a();\n\nb();", "a();\n\nb();");
         bt('var a = 1 var b = 2', "var a = 1\nvar b = 2");
         bt('var a=1, b=c[d], e=6;', 'var a = 1,\n    b = c[d],\n    e = 6;');
+        bt('let a = 1 let b = 2', "let a = 1\nlet b = 2");
+        bt('let a=1, b=c[d], e=6;', 'let a = 1,\n    b = c[d],\n    e = 6;');
+        bt('const a = 1 const b = 2', "const a = 1\nconst b = 2");
+        bt('const a=1, b=c[d], e=6;', 'const a = 1,\n    b = c[d],\n    e = 6;');
         bt('a = " 12345 "');
         bt("a = ' 12345 '");
         bt('if (a == 1) b = 2;', "if (a == 1) b = 2;");
@@ -72,6 +81,7 @@ class TestJSBeautifier(unittest.TestCase):
         bt('a.b({c:"d"})', 'a.b({\n    c: "d"\n})');
         bt('a.b\n(\n{\nc:\n"d"\n}\n)', 'a.b({\n    c: "d"\n})');
         bt('a=!b', 'a = !b');
+        bt('a=!!b', 'a = !!b');
         bt('a?b:c', 'a ? b : c');
         bt('a?1:2', 'a ? 1 : 2');
         bt('a?(b):c', 'a ? (b) : c');
@@ -117,6 +127,7 @@ class TestJSBeautifier(unittest.TestCase):
         bt('if (template.user[n] in bk) foo();');
         bt('{{}/z/}', "{\n    {}\n    /z/\n}");
         bt('return 45', "return 45");
+        bt('return this.prevObject ||\n\n    this.constructor(null);');
         bt('If[1]', "If[1]");
         bt('Then[1]', "Then[1]");
         bt('a = 1e10', "a = 1e10");
@@ -161,10 +172,20 @@ class TestJSBeautifier(unittest.TestCase):
         bt("function x(){(a||b).c()}", "function x() {\n    (a || b).c()\n}");
         bt("function x(){return - 1}", "function x() {\n    return -1\n}");
         bt("function x(){return ! a}", "function x() {\n    return !a\n}");
+        bt("x => x", "x => x");
+        bt("(x) => x", "(x) => x");
+        bt("x => { x }", "x => {\n    x\n}");
+        bt("(x) => { x }", "(x) => {\n    x\n}");
 
         # a common snippet in jQuery plugins
         bt("settings = $.extend({},defaults,settings);", "settings = $.extend({}, defaults, settings);");
 
+        # reserved words used as property names
+        bt("$http().then().finally().default()", "$http().then().finally().default()");
+        bt("$http()\n.then()\n.finally()\n.default()", "$http()\n    .then()\n    .finally()\n    .default()");
+        bt("$http().when.in.new.catch().throw()", "$http().when.in.new.catch().throw()");
+        bt("$http()\n.when\n.in\n.new\n.catch()\n.throw()", "$http()\n    .when\n    .in\n    .new\n    .catch()\n    .throw()");
+
         bt('{xxx;}()', '{\n    xxx;\n}()');
 
         bt("a = 'a'\nb = 'b'");
@@ -189,6 +210,7 @@ class TestJSBeautifier(unittest.TestCase):
         test_fragment('"incomplete-string');
         test_fragment("'incomplete-string");
         test_fragment('/incomplete-regex');
+        test_fragment('`incomplete-regex');
 
         test_fragment('{a:1},{a:2}', '{\n    a: 1\n}, {\n    a: 2\n}');
         test_fragment('var ary=[{a:1}, {a:2}];', 'var ary = [{\n    a: 1\n}, {\n    a: 2\n}];');
@@ -243,6 +265,7 @@ class TestJSBeautifier(unittest.TestCase):
         # allow unescaped / in char classes
         bt('a(/[a/b]/);b()', "a(/[a/b]/);\nb()");
 
+        bt('function foo() {\n    return [\n        "one",\n        "two"\n    ];\n}');
         bt('a=[[1,2],[4,5],[7,8]]', "a = [\n    [1, 2],\n    [4, 5],\n    [7, 8]\n]");
         bt('a=[[1,2],[4,5],function(){},[7,8]]',
             "a = [\n    [1, 2],\n    [4, 5],\n    function() {},\n    [7, 8]\n]");
@@ -269,9 +292,9 @@ class TestJSBeautifier(unittest.TestCase):
         bt('catch(e)', 'catch (e)');
 
         bt('var a=1,b={foo:2,bar:3},{baz:4,wham:5},c=4;',
-           'var a = 1,\n    b = {\n        foo: 2,\n        bar: 3\n    }, {\n        baz: 4,\n        wham: 5\n    }, c = 4;');
+            'var a = 1,\n    b = {\n        foo: 2,\n        bar: 3\n    },\n    {\n        baz: 4,\n        wham: 5\n    }, c = 4;');
         bt('var a=1,b={foo:2,bar:3},{baz:4,wham:5},\nc=4;',
-           'var a = 1,\n    b = {\n        foo: 2,\n        bar: 3\n    }, {\n        baz: 4,\n        wham: 5\n    },\n    c = 4;');
+            'var a = 1,\n    b = {\n        foo: 2,\n        bar: 3\n    },\n    {\n        baz: 4,\n        wham: 5\n    },\n    c = 4;');
 
         # inline comment
         bt('function x(/*int*/ start, /*string*/ foo)', 'function x( /*int*/ start, /*string*/ foo)');
@@ -298,12 +321,16 @@ class TestJSBeautifier(unittest.TestCase):
         bt('if (a) a()\nnewline()');
         bt('a=typeof(x)', 'a = typeof(x)');
 
-        bt('var a = function() {\n    return null;\n},\n    b = false;');
+        bt('var a = function() {\n        return null;\n    },\n    b = false;');
 
         bt('var a = function() {\n    func1()\n}');
         bt('var a = function() {\n    func1()\n}\nvar b = function() {\n    func2()\n}');
 
-
+        # code with and without semicolons
+        bt( 'var whatever = require("whatever");\nfunction() {\n    a = 6;\n}',
+            'var whatever = require("whatever");\n\nfunction() {\n    a = 6;\n}');
+        bt( 'var whatever = require("whatever")\nfunction() {\n    a = 6\n}',
+            'var whatever = require("whatever")\n\nfunction() {\n    a = 6\n}');
 
 
         self.options.jslint_happy = True
@@ -325,7 +352,7 @@ class TestJSBeautifier(unittest.TestCase):
         bt('switch(x){case -1:break;case !y:break;}',
             'switch (x) {\n    case -1:\n        break;\n    case !y:\n        break;\n}');
         test_fragment("// comment 2\n(function()", "// comment 2\n(function()"); # typical greasemonkey start
-        bt("var a2, b2, c2, d2 = 0, c = function() {}, d = '';", "var a2, b2, c2, d2 = 0,\n    c = function() {}, d = '';");
+        bt("var a2, b2, c2, d2 = 0, c = function() {}, d = '';", "var a2, b2, c2, d2 = 0,\n    c = function() {},\n    d = '';");
         bt("var a2, b2, c2, d2 = 0, c = function() {},\nd = '';", "var a2, b2, c2, d2 = 0,\n    c = function() {},\n    d = '';");
         bt('var o2=$.extend(a);function(){alert(x);}', 'var o2 = $.extend(a);\n\nfunction() {\n    alert(x);\n}');
 
@@ -384,7 +411,7 @@ class TestJSBeautifier(unittest.TestCase):
 
 
         self.options.preserve_newlines = True;
-        bt('var\na=do_preserve_newlines;', 'var\na = do_preserve_newlines;')
+        bt('var\na=do_preserve_newlines;', 'var\n    a = do_preserve_newlines;')
         bt('// a\n// b\n\n// c\n// d')
         bt('if (foo) //  comment\n{\n    bar();\n}')
 
@@ -408,8 +435,69 @@ class TestJSBeautifier(unittest.TestCase):
             '{\n    a([\n        [a1]\n    ], {\n        b;\n    });\n}');
         bt("a();\n   [\n   ['sdfsdfsd'],\n        ['sdfsdfsdf']\n   ].toString();",
             "a();\n[\n    ['sdfsdfsd'],\n    ['sdfsdfsdf']\n].toString();");
+        bt("a();\na = [\n   ['sdfsdfsd'],\n        ['sdfsdfsdf']\n   ].toString();",
+            "a();\na = [\n    ['sdfsdfsd'],\n    ['sdfsdfsdf']\n].toString();");
         bt("function() {\n    Foo([\n        ['sdfsdfsd'],\n        ['sdfsdfsdf']\n    ]);\n}",
             "function() {\n    Foo([\n        ['sdfsdfsd'],\n        ['sdfsdfsdf']\n    ]);\n}");
+        bt('function foo() {\n    return [\n        "one",\n        "two"\n    ];\n}');
+        # 4 spaces per indent input, processed with 4-spaces per indent
+        bt( "function foo() {\n" +
+            "    return [\n" +
+            "        {\n" +
+            "            one: 'x',\n" +
+            "            two: [\n" +
+            "                {\n" +
+            "                    id: 'a',\n" +
+            "                    name: 'apple'\n" +
+            "                }, {\n" +
+            "                    id: 'b',\n" +
+            "                    name: 'banana'\n" +
+            "                }\n" +
+            "            ]\n" +
+            "        }\n" +
+            "    ];\n" +
+            "}",
+            "function foo() {\n" +
+            "    return [{\n" +
+            "        one: 'x',\n" +
+            "        two: [{\n" +
+            "            id: 'a',\n" +
+            "            name: 'apple'\n" +
+            "        }, {\n" +
+            "            id: 'b',\n" +
+            "            name: 'banana'\n" +
+            "        }]\n" +
+            "    }];\n" +
+            "}");
+        # 3 spaces per indent input, processed with 4-spaces per indent
+        bt( "function foo() {\n" +
+            "   return [\n" +
+            "      {\n" +
+            "         one: 'x',\n" +
+            "         two: [\n" +
+            "            {\n" +
+            "               id: 'a',\n" +
+            "               name: 'apple'\n" +
+            "            }, {\n" +
+            "               id: 'b',\n" +
+            "               name: 'banana'\n" +
+            "            }\n" +
+            "         ]\n" +
+            "      }\n" +
+            "   ];\n" +
+            "}",
+            "function foo() {\n" +
+            "    return [{\n" +
+            "        one: 'x',\n" +
+            "        two: [{\n" +
+            "            id: 'a',\n" +
+            "            name: 'apple'\n" +
+            "        }, {\n" +
+            "            id: 'b',\n" +
+            "            name: 'banana'\n" +
+            "        }]\n" +
+            "    }];\n" +
+            "}");
 
         self.options.keep_array_indentation = True;
         bt("a = ['a', 'b', 'c',\n    'd', 'e', 'f']");
@@ -426,8 +514,47 @@ class TestJSBeautifier(unittest.TestCase):
             '{\n    a([[a1]], {\n        b;\n    });\n}');
         bt("a();\n   [\n   ['sdfsdfsd'],\n        ['sdfsdfsdf']\n   ].toString();",
             "a();\n   [\n   ['sdfsdfsd'],\n        ['sdfsdfsdf']\n   ].toString();");
+        bt("a();\na = [\n   ['sdfsdfsd'],\n        ['sdfsdfsdf']\n   ].toString();",
+            "a();\na = [\n   ['sdfsdfsd'],\n        ['sdfsdfsdf']\n   ].toString();");
         bt("function() {\n    Foo([\n        ['sdfsdfsd'],\n        ['sdfsdfsdf']\n    ]);\n}",
             "function() {\n    Foo([\n        ['sdfsdfsd'],\n        ['sdfsdfsdf']\n    ]);\n}");
+        bt('function foo() {\n    return [\n        "one",\n        "two"\n    ];\n}');
+        # 4 spaces per indent input, processed with 4-spaces per indent
+        bt( "function foo() {\n" +
+            "    return [\n" +
+            "        {\n" +
+            "            one: 'x',\n" +
+            "            two: [\n" +
+            "                {\n" +
+            "                    id: 'a',\n" +
+            "                    name: 'apple'\n" +
+            "                }, {\n" +
+            "                    id: 'b',\n" +
+            "                    name: 'banana'\n" +
+            "                }\n" +
+            "            ]\n" +
+            "        }\n" +
+            "    ];\n" +
+            "}");
+        # 3 spaces per indent input, processed with 4-spaces per indent
+        # Should be unchanged, but is not - #445
+#         bt( "function foo() {\n" +
+#             "   return [\n" +
+#             "      {\n" +
+#             "         one: 'x',\n" +
+#             "         two: [\n" +
+#             "            {\n" +
+#             "               id: 'a',\n" +
+#             "               name: 'apple'\n" +
+#             "            }, {\n" +
+#             "               id: 'b',\n" +
+#             "               name: 'banana'\n" +
+#             "            }\n" +
+#             "         ]\n" +
+#             "      }\n" +
+#             "   ];\n" +
+#             "}");
+
         self.options.keep_array_indentation = False;
 
         bt('a = //comment\n/regex/;');
@@ -448,6 +575,8 @@ class TestJSBeautifier(unittest.TestCase):
             "try\n{\n    a();\n}\ncatch (b)\n{\n    c();\n}\ncatch (d)\n{}\nfinally\n{\n    e();\n}");
         bt('if(a){b();}else if(c) foo();',
             "if (a)\n{\n    b();\n}\nelse if (c) foo();");
+        bt('if(X)if(Y)a();else b();else c();',
+            "if (X)\n    if (Y) a();\n    else b();\nelse c();");
         bt("if (a) {\n// comment\n}else{\n// comment\n}",
             "if (a)\n{\n    // comment\n}\nelse\n{\n    // comment\n}"); # if/else statement with empty body
         bt('if (x) {y} else { if (x) {y}}',
@@ -473,7 +602,9 @@ class TestJSBeautifier(unittest.TestCase):
         bt('a: do {} while (); xxx', 'a: do {} while ();\nxxx');
         bt('var a = new function();');
         bt('var a = new function() {};');
-        bt('var a = new function a()\n    {};');
+        bt('var a = new function()\n{};', 'var a = new function() {};');
+        bt('var a = new function a()\n{};');
+        bt('var a = new function a()\n    {},\n    b = new function b()\n    {};');
         test_fragment('new function');
         bt("foo({\n    'a': 1\n},\n10);",
             "foo(\n    {\n        'a': 1\n    },\n    10);");
@@ -743,6 +874,7 @@ class TestJSBeautifier(unittest.TestCase):
         bt("{\n    var a = set\n    foo();\n}")
         bt("var x = {\n    get function()\n}")
         bt("var x = {\n    set function()\n}")
+        bt("var x = set\n\na() {}", "var x = set\n\n    a() {}");
         bt("var x = set\n\nfunction() {}", "var x = set\n\n    function() {}")
 
         bt('<!-- foo\nbar();\n-->')
@@ -807,55 +939,65 @@ class TestJSBeautifier(unittest.TestCase):
         self.options.break_chained_methods = False
         self.options.preserve_newlines = False
 
-        self.options.preserve_newlines = False
-        self.options.wrap_line_length = 0
+
+        # Line wrap test intputs
         #..............---------1---------2---------3---------4---------5---------6---------7
         #..............1234567890123456789012345678901234567890123456789012345678901234567890
-        test_fragment('foo.bar().baz().cucumber((fat && "sassy") || (leans\n&& mean));\n' +
+        wrap_input_1=('foo.bar().baz().cucumber((fat && "sassy") || (leans\n&& mean));\n' +
                       'Test_very_long_variable_name_this_should_never_wrap\n.but_this_can\n' +
                       'if (wraps_can_occur && inside_an_if_block) that_is_\n.okay();\n' +
                       'object_literal = {\n' +
                       '    property: first_token_should_never_wrap + but_this_can,\n' +
+                      '    propertz: first_token_should_never_wrap + !but_this_can,\n' +
                       '    proper: "first_token_should_never_wrap" + "but_this_can"\n' +
-                      '}',
+                      '}')
+
+        #..............---------1---------2---------3---------4---------5---------6---------7
+        #..............1234567890123456789012345678901234567890123456789012345678901234567890
+        wrap_input_2=('{\n' +
+                      '    foo.bar().baz().cucumber((fat && "sassy") || (leans\n&& mean));\n' +
+                      '    Test_very_long_variable_name_this_should_never_wrap\n.but_this_can\n' +
+                      '    if (wraps_can_occur && inside_an_if_block) that_is_\n.okay();\n' +
+                      '    object_literal = {\n' +
+                      '        property: first_token_should_never_wrap + but_this_can,\n' +
+                      '        propertz: first_token_should_never_wrap + !but_this_can,\n' +
+                      '        proper: "first_token_should_never_wrap" + "but_this_can"\n' +
+                      '    }' +
+                      '}')
+
+        self.options.preserve_newlines = False
+        self.options.wrap_line_length = 0
+        #..............---------1---------2---------3---------4---------5---------6---------7
+        #..............1234567890123456789012345678901234567890123456789012345678901234567890
+        test_fragment(wrap_input_1,
                       # expected #
                       'foo.bar().baz().cucumber((fat && "sassy") || (leans && mean));\n' +
                       'Test_very_long_variable_name_this_should_never_wrap.but_this_can\n' +
                       'if (wraps_can_occur && inside_an_if_block) that_is_.okay();\n' +
                       'object_literal = {\n' +
                       '    property: first_token_should_never_wrap + but_this_can,\n' +
+                      '    propertz: first_token_should_never_wrap + !but_this_can,\n' +
                       '    proper: "first_token_should_never_wrap" + "but_this_can"\n' +
                       '}');
 
         self.options.wrap_line_length = 70
         #..............---------1---------2---------3---------4---------5---------6---------7
         #..............1234567890123456789012345678901234567890123456789012345678901234567890
-        test_fragment('foo.bar().baz().cucumber((fat && "sassy") || (leans\n&& mean));\n' +
-                      'Test_very_long_variable_name_this_should_never_wrap\n.but_this_can\n' +
-                      'if (wraps_can_occur && inside_an_if_block) that_is_\n.okay();\n' +
-                      'object_literal = {\n' +
-                      '    property: first_token_should_never_wrap + but_this_can,\n' +
-                      '    proper: "first_token_should_never_wrap" + "but_this_can"\n' +
-                      '}',
+        test_fragment(wrap_input_1,
                       # expected #
                       'foo.bar().baz().cucumber((fat && "sassy") || (leans && mean));\n' +
                       'Test_very_long_variable_name_this_should_never_wrap.but_this_can\n' +
                       'if (wraps_can_occur && inside_an_if_block) that_is_.okay();\n' +
                       'object_literal = {\n' +
                       '    property: first_token_should_never_wrap + but_this_can,\n' +
+                      '    propertz: first_token_should_never_wrap + !but_this_can,\n' +
                       '    proper: "first_token_should_never_wrap" + "but_this_can"\n' +
                       '}');
 
         self.options.wrap_line_length = 40
         #..............---------1---------2---------3---------4---------5---------6---------7
         #..............1234567890123456789012345678901234567890123456789012345678901234567890
-        test_fragment('foo.bar().baz().cucumber((fat && "sassy") || (leans\n&& mean));\n' +
-                      'Test_very_long_variable_name_this_should_never_wrap\n.but_this_can\n' +
-                      'if (wraps_can_occur && inside_an_if_block) that_is_\n.okay();\n' +
-                      'object_literal = {\n' +
-                      '    property: first_token_should_never_wrap + but_this_can,\n' +
-                      '    proper: "first_token_should_never_wrap" + "but_this_can"\n' +
-                      '}',
+        test_fragment(wrap_input_1,
                       # expected #
                       'foo.bar().baz().cucumber((fat &&\n' +
                       '    "sassy") || (leans && mean));\n' +
@@ -866,6 +1008,8 @@ class TestJSBeautifier(unittest.TestCase):
                       'object_literal = {\n' +
                       '    property: first_token_should_never_wrap +\n' +
                       '        but_this_can,\n' +
+                      '    propertz: first_token_should_never_wrap +\n' +
+                      '        !but_this_can,\n' +
                       '    proper: "first_token_should_never_wrap" +\n' +
                       '        "but_this_can"\n' +
                       '}');
@@ -874,13 +1018,7 @@ class TestJSBeautifier(unittest.TestCase):
         # NOTE: wrap is only best effort - line continues until next wrap point is found.
         #..............---------1---------2---------3---------4---------5---------6---------7
         #..............1234567890123456789012345678901234567890123456789012345678901234567890
-        test_fragment('foo.bar().baz().cucumber((fat && "sassy") || (leans\n&& mean));\n' +
-                      'Test_very_long_variable_name_this_should_never_wrap\n.but_this_can\n' +
-                      'if (wraps_can_occur && inside_an_if_block) that_is_\n.okay();\n' +
-                      'object_literal = {\n' +
-                      '    property: first_token_should_never_wrap + but_this_can,\n' +
-                      '    proper: "first_token_should_never_wrap" + "but_this_can"\n' +
-                      '}',
+        test_fragment(wrap_input_1,
                       # expected #
                       'foo.bar().baz().cucumber((fat && "sassy") ||\n' +
                       '    (leans && mean));\n' +
@@ -891,6 +1029,8 @@ class TestJSBeautifier(unittest.TestCase):
                       'object_literal = {\n' +
                       '    property: first_token_should_never_wrap +\n' +
                       '        but_this_can,\n' +
+                      '    propertz: first_token_should_never_wrap +\n' +
+                      '        !but_this_can,\n' +
                       '    proper: "first_token_should_never_wrap" +\n' +
                       '        "but_this_can"\n' +
                       '}');
@@ -900,15 +1040,7 @@ class TestJSBeautifier(unittest.TestCase):
         # NOTE: wrap is only best effort - line continues until next wrap point is found.
         #..............---------1---------2---------3---------4---------5---------6---------7
         #..............1234567890123456789012345678901234567890123456789012345678901234567890
-        test_fragment('{\n' +
-                      '    foo.bar().baz().cucumber((fat && "sassy") || (leans\n&& mean));\n' +
-                      '    Test_very_long_variable_name_this_should_never_wrap\n.but_this_can\n' +
-                      '    if (wraps_can_occur && inside_an_if_block) that_is_\n.okay();\n' +
-                      '    object_literal = {\n' +
-                      '        property: first_token_should_never_wrap + but_this_can,\n' +
-                      '        proper: "first_token_should_never_wrap" + "but_this_can"\n' +
-                      '    }' +
-                      '}',
+        test_fragment(wrap_input_2,
                       # expected #
                       '{\n' +
                       '    foo.bar().baz().cucumber((fat && "sassy") ||\n' +
@@ -920,6 +1052,8 @@ class TestJSBeautifier(unittest.TestCase):
                       '    object_literal = {\n' +
                       '        property: first_token_should_never_wrap +\n' +
                       '            but_this_can,\n' +
+                      '        propertz: first_token_should_never_wrap +\n' +
+                      '            !but_this_can,\n' +
                       '        proper: "first_token_should_never_wrap" +\n' +
                       '            "but_this_can"\n' +
                       '    }\n'+
@@ -929,13 +1063,7 @@ class TestJSBeautifier(unittest.TestCase):
         self.options.wrap_line_length = 0
         #..............---------1---------2---------3---------4---------5---------6---------7
         #..............1234567890123456789012345678901234567890123456789012345678901234567890
-        test_fragment('foo.bar().baz().cucumber((fat && "sassy") || (leans\n&& mean));\n' +
-                      'Test_very_long_variable_name_this_should_never_wrap\n.but_this_can\n' +
-                      'if (wraps_can_occur && inside_an_if_block) that_is_\n.okay();\n' +
-                      'object_literal = {\n' +
-                      '    property: first_token_should_never_wrap + but_this_can,\n' +
-                      '    proper: "first_token_should_never_wrap" + "but_this_can"\n' +
-                      '}',
+        test_fragment(wrap_input_1,
                       # expected #
                       'foo.bar().baz().cucumber((fat && "sassy") || (leans && mean));\n' +
                       'Test_very_long_variable_name_this_should_never_wrap\n' +
@@ -944,6 +1072,7 @@ class TestJSBeautifier(unittest.TestCase):
                       '    .okay();\n' +
                       'object_literal = {\n' +
                       '    property: first_token_should_never_wrap + but_this_can,\n' +
+                      '    propertz: first_token_should_never_wrap + !but_this_can,\n' +
                       '    proper: "first_token_should_never_wrap" + "but_this_can"\n' +
                       '}');
 
@@ -951,13 +1080,7 @@ class TestJSBeautifier(unittest.TestCase):
         self.options.wrap_line_length = 70
         #..............---------1---------2---------3---------4---------5---------6---------7
         #..............1234567890123456789012345678901234567890123456789012345678901234567890
-        test_fragment('foo.bar().baz().cucumber((fat && "sassy") || (leans\n&& mean));\n' +
-                      'Test_very_long_variable_name_this_should_never_wrap\n.but_this_can\n' +
-                      'if (wraps_can_occur && inside_an_if_block) that_is_\n.okay();\n' +
-                      'object_literal = {\n' +
-                      '    property: first_token_should_never_wrap + but_this_can,\n' +
-                      '    proper: "first_token_should_never_wrap" + "but_this_can"\n' +
-                      '}',
+        test_fragment(wrap_input_1,
                       # expected #
                       'foo.bar().baz().cucumber((fat && "sassy") || (leans && mean));\n' +
                       'Test_very_long_variable_name_this_should_never_wrap\n' +
@@ -966,6 +1089,7 @@ class TestJSBeautifier(unittest.TestCase):
                       '    .okay();\n' +
                       'object_literal = {\n' +
                       '    property: first_token_should_never_wrap + but_this_can,\n' +
+                      '    propertz: first_token_should_never_wrap + !but_this_can,\n' +
                       '    proper: "first_token_should_never_wrap" + "but_this_can"\n' +
                       '}');
 
@@ -973,13 +1097,7 @@ class TestJSBeautifier(unittest.TestCase):
         self.options.wrap_line_length = 40
         #..............---------1---------2---------3---------4---------5---------6---------7
         #..............1234567890123456789012345678901234567890123456789012345678901234567890
-        test_fragment('foo.bar().baz().cucumber((fat && "sassy") || (leans\n&& mean));\n' +
-                      'Test_very_long_variable_name_this_should_never_wrap\n.but_this_can\n' +
-                      'if (wraps_can_occur && inside_an_if_block) that_is_\n.okay();\n' +
-                      'object_literal = {\n' +
-                      '    property: first_token_should_never_wrap + but_this_can,\n' +
-                      '    proper: "first_token_should_never_wrap" + "but_this_can"\n' +
-                      '}',
+        test_fragment(wrap_input_1,
                       # expected #
                       'foo.bar().baz().cucumber((fat &&\n' +
                       '    "sassy") || (leans && mean));\n' +
@@ -991,6 +1109,8 @@ class TestJSBeautifier(unittest.TestCase):
                       'object_literal = {\n' +
                       '    property: first_token_should_never_wrap +\n' +
                       '        but_this_can,\n' +
+                      '    propertz: first_token_should_never_wrap +\n' +
+                      '        !but_this_can,\n' +
                       '    proper: "first_token_should_never_wrap" +\n' +
                       '        "but_this_can"\n' +
                       '}');
@@ -999,13 +1119,7 @@ class TestJSBeautifier(unittest.TestCase):
         # NOTE: wrap is only best effort - line continues until next wrap point is found.
         #..............---------1---------2---------3---------4---------5---------6---------7
         #..............1234567890123456789012345678901234567890123456789012345678901234567890
-        test_fragment('foo.bar().baz().cucumber((fat && "sassy") || (leans\n&& mean));\n' +
-                      'Test_very_long_variable_name_this_should_never_wrap\n.but_this_can\n' +
-                      'if (wraps_can_occur && inside_an_if_block) that_is_\n.okay();\n' +
-                      'object_literal = {\n' +
-                      '    property: first_token_should_never_wrap + but_this_can,\n' +
-                      '    proper: "first_token_should_never_wrap" + "but_this_can"\n' +
-                      '}',
+        test_fragment(wrap_input_1,
                       # expected #
                       'foo.bar().baz().cucumber((fat && "sassy") ||\n' +
                       '    (leans && mean));\n' +
@@ -1017,6 +1131,8 @@ class TestJSBeautifier(unittest.TestCase):
                       'object_literal = {\n' +
                       '    property: first_token_should_never_wrap +\n' +
                       '        but_this_can,\n' +
+                      '    propertz: first_token_should_never_wrap +\n' +
+                      '        !but_this_can,\n' +
                       '    proper: "first_token_should_never_wrap" +\n' +
                       '        "but_this_can"\n' +
                       '}');
@@ -1025,15 +1141,7 @@ class TestJSBeautifier(unittest.TestCase):
         # NOTE: wrap is only best effort - line continues until next wrap point is found.
         #..............---------1---------2---------3---------4---------5---------6---------7
         #..............1234567890123456789012345678901234567890123456789012345678901234567890
-        test_fragment('{\n' +
-                      '    foo.bar().baz().cucumber((fat && "sassy") || (leans\n&& mean));\n' +
-                      '    Test_very_long_variable_name_this_should_never_wrap\n.but_this_can\n' +
-                      '    if (wraps_can_occur && inside_an_if_block) that_is_\n.okay();\n' +
-                      '    object_literal = {\n' +
-                      '        property: first_token_should_never_wrap + but_this_can,\n' +
-                      '        proper: "first_token_should_never_wrap" + "but_this_can"\n' +
-                      '    }' +
-                      '}',
+        test_fragment(wrap_input_2,
                       # expected #
                       '{\n' +
                       '    foo.bar().baz().cucumber((fat && "sassy") ||\n' +
@@ -1046,6 +1154,8 @@ class TestJSBeautifier(unittest.TestCase):
                       '    object_literal = {\n' +
                       '        property: first_token_should_never_wrap +\n' +
                       '            but_this_can,\n' +
+                      '        propertz: first_token_should_never_wrap +\n' +
+                      '            !but_this_can,\n' +
                       '        proper: "first_token_should_never_wrap" +\n' +
                       '            "but_this_can"\n' +
                       '    }\n'+
@@ -1067,6 +1177,8 @@ class TestJSBeautifier(unittest.TestCase):
         bt('foo = {\n    x: y, // #44\n    w: z // #44\n}');
         bt('switch (x) {\n    case "a":\n        // comment on newline\n        break;\n    case "b": // comment on same line\n        break;\n}');
 
+        bt('if (true ||\n!true) return;', 'if (true || !true) return;');
+
         # these aren't ready yet.
         #bt('if (foo) // comment\n    bar() /*i*/ + baz() /*j\n*/ + asdf();');
         bt('if\n(foo)\nif\n(bar)\nif\n(baz)\nwhee();\na();',
@@ -1149,6 +1261,8 @@ class TestJSBeautifier(unittest.TestCase):
         bt('foo = {\n    x: y, // #44\n    w: z // #44\n}');
         bt('switch (x) {\n    case "a":\n        // comment on newline\n        break;\n    case "b": // comment on same line\n        break;\n}');
 
+        bt('if (true ||\n!true) return;', 'if (true ||\n    !true) return;');
+
         # these aren't ready yet.
         # bt('if (foo) // comment\n    bar() /*i*/ + baz() /*j\n*/ + asdf();');
         bt('if\n(foo)\nif\n(bar)\nif\n(baz)\nwhee();\na();',
@@ -1215,6 +1329,7 @@ class TestJSBeautifier(unittest.TestCase):
 
         # Test the option to have spaces within parens
         self.options.space_in_paren = False
+        self.options.space_in_empty_paren = False
         bt('if(p) foo(a,b)', 'if (p) foo(a, b)');
         bt('try{while(true){willThrow()}}catch(result)switch(result){case 1:++result }',
            'try {\n    while (true) {\n        willThrow()\n    }\n} catch (result) switch (result) {\n    case 1:\n        ++result\n}');
@@ -1242,8 +1357,26 @@ class TestJSBeautifier(unittest.TestCase):
         bt('a= f[b];',
             'a = f[ b ];');
 
-        self.options.space_in_paren = False
+        self.options.space_in_empty_paren = True
+        bt('if(p) foo(a,b)', 'if ( p ) foo( a, b )');
+        bt('try{while(true){willThrow()}}catch(result)switch(result){case 1:++result }',
+           'try {\n    while ( true ) {\n        willThrow( )\n    }\n} catch ( result ) switch ( result ) {\n    case 1:\n        ++result\n}');
+        bt('((e/((a+(b)*c)-d))^2)*5;', '( ( e / ( ( a + ( b ) * c ) - d ) ) ^ 2 ) * 5;');
+        bt('function f(a,b) {if(a) b()}function g(a,b) {if(!a) b()}',
+            'function f( a, b ) {\n    if ( a ) b( )\n}\n\nfunction g( a, b ) {\n    if ( !a ) b( )\n}');
+        bt('a=[ ];',
+            'a = [ ];');
+        bt('a=[b,c,d];',
+            'a = [ b, c, d ];');
+        bt('a= f[b];',
+            'a = f[ b ];');
 
+        self.options.space_in_paren = False
+        self.options.space_in_empty_paren = False
+
+        # Test template strings
+        bt('`This is a ${template} string.`', '`This is a ${template} string.`');
+        bt('`This\n  is\n  a\n  ${template}\n  string.`', '`This\n  is\n  a\n  ${template}\n  string.`');
 
         # Test that e4x literals passed through when e4x-option is enabled
         bt('xml=<a b="c"><d/><e>\n foo</e>x</a>;', 'xml = < a b = "c" > < d / > < e >\n    foo < /e>x</a > ;');
@@ -1351,12 +1484,10 @@ class TestJSBeautifier(unittest.TestCase):
            '    });');
         # END tests for issue 281
 
-        # This is what I think these should look like related #256
-        # we don't have the ability yet
-        #bt('var a=1,b={bang:2},c=3;',
-        #   'var a = 1,\n    b = {\n        bang: 2\n    },\n     c = 3;');
-        #bt('var a={bing:1},b=2,c=3;',
-        #   'var a = {\n        bing: 1\n    },\n    b = 2,\n    c = 3;');
+        bt('var a=1,b={bang:2},c=3;',
+            'var a = 1,\n    b = {\n        bang: 2\n    },\n    c = 3;');
+        bt('var a={bing:1},b=2,c=3;',
+            'var a = {\n        bing: 1\n    },\n    b = 2,\n    c = 3;');
 
 
 
